from urlparse import urlparse
from .page import Page
import re, json
import requests
from .storage_editor import StorageEditor
from .page_properties import PagePropertiesEditor

import logging
logger = logging.getLogger('confluence.api')
#logger.setLevel(logging.DEBUG)

filter_func = filter

def is_string(s):
    # python 2.7
    return isinstance(s, basestring)

class ConfluenceError(StandardError):
    pass

class ConfluenceAPI:
    def __init__(self, config):
        self.config = config
        self.hostname = urlparse(config['baseurl']).hostname

    def set_args(self, args):
        self.args = args

    def get_args(self):
        return self.args

    def _getauth(self):
        '''return tuple (user, password) for accessing confluence'''
        import netrc

        if self.config.get('username'):

            if not self.config['password']:
                import keyring
                baseurl = self.config['baseurl']
                password = keyring.get_password(baseurl, self.config['username'])
                if password is None:
                    import getpass
                    password = getpass

            else:
                password = self.config.get('password')

            if password is not None:
                return (self.config['username'], password)

        logger.info('hostname: %s', self.hostname)
        (user, account, password) = netrc.netrc().authenticators(self.hostname)
        return (user, password)

    def __getattr__(self, name):
        if name == 'session':
            self.session = requests.Session()
            self.session.auth = self._getauth()
            return self.session

        raise AttributeError(name)

    def request(self, method, endpoint, params=None, stream=None, **kwargs):
        url = self.config['baseurl'] + endpoint
        if params is None:
            params = {}

        headers = { 'X-Atlassian-Token': 'no-check' }
        if isinstance(params, dict):
            params.update(kwargs)

        try:

            if method == 'GET':
                response = self.session.get(url, params=params, headers=headers, stream=stream)
            else:
                headers.update({'Content-Type': 'application/json', 'Accept':'application/json'})
                response = self.session.request(method, url, data=json.dumps(params), headers=headers)

        except StandardError as e:
            logger.info("error in request %s %s with params %s", method, endpoint, params)
            raise

        if response.status_code >= 400:
            self.session.close()
            del self.session

            error = ''
            logger.info("error: %s %s, %s: %s", method, url, params, response.text)

            raise ConfluenceError(response.text)

        if not stream:
            if response.text:
                return response.json()
        else:
            return response

        return {}

    def get(self, endpoint, params=None, **kwargs):
        return self.request('GET', endpoint, params, **kwargs)

    def put(self, endpoint, params=None, **kwargs):
        return self.request('PUT', endpoint, params, **kwargs)

    def post(self, endpoint, params=None, **kwargs):
        return self.request('POST', endpoint, params, **kwargs)

    def delete(self, endpoint, params=None, **kwargs):
        return self.request('DELETE', endpoint, params, **kwargs)

    def createSpace(self, key, name, description=''):
        return self.post( '/rest/api/space',
            key=key, name=name, type='global', description={
                'plain': {
                    'value': description,
                    'representation': 'plain'
                }
            })

    def copySpace(self, source_key, key, name, description=''):
        source_space = self.getSpace(source_key)

        # create new space
        space = self.createSpace(key, name, description)
        target_space = self.getSpace(key)
        #log.info("Created new ps")

        self.copyPage(source_space['_expandable']['homepage'], target_space['_expandable']['homepage'])

        # self.updatePage(target_page['id'], version=target_page['version']['number'], title=target_page['title'], storage=source_page['body']['storage']['value'])
        #
        # subpages = self.getChildren(source_page['id'], type='page')
        # # now create page tree from source page
        # for subpage in subpages:
        #     sp = self.getPage(subpage['id'], expand='body.storage')
        #     self.createPage(target_space['key'], subpage['title'], storage = sp['body']['storage']['value'], parent=target_page['id'])

    def getUser(self, username, expand=''):
        """get user information"""
        return self.get("/rest/api/user", username=username, expand=expand)

    def copyPage(self, source, target=None, recursive=True, parent=None, space=None, delete=False):
        '''copy source page as child of target and descend all children

        :param source:
           content url or a pageid
        :param recursive:
           (default True) copy children
        :param parent:
           parent for current target
        :param target:
           where to copy a page
        :param space:
           where to copy a page
        :param delete:
           delete children not present in source
        '''

        source_page = self.getPage(source, expand='body.storage,space')
        target_page = self.getPage(target, expand='version,space')

        if target_page is None:
            target_page = self.createPage(
                space   = space,
                title   = target,
                storage = source_page['body']['storage']['value'],
                parent  = parent
            )
            logger.info("Create Page: %s, %s", space, target)
        else:
            self.updatePage( target_page['id'],
                version = target_page['version']['number'],
                title   = target_page['title'],
                storage = source_page['body']['storage']['value']
            )
            logger.info("Update Page: %s, %s", space, target_page['title'])

        if recursive:
            source_subpages = sorted([ (p['title'], p) for p in self.getChildren(source_page['id'], type='page') ])
            target_subpages = sorted([ (p['title'], p) for p in self.getChildren(target_page['id'], type='page') ])

            # TODO: assert that there are no duplicate page titles

            source_subpage = dict(source_subpages)
            target_subpage = dict(target_subpages)

            assert len(source_subpage.keys()) == len(source_subpages)
            assert len(target_subpage.keys()) == len(target_subpages)

            for title, page in source_subpage.items():
                if title in target_subpage:
                    subpage = target_subpage[title]['id']
                else:
                    subpage = title

                self.copyPage(page['id'], target=subpage, recursive=recursive,
                    parent=target_page['id'], space=space)

            if delete:
                for title, page in target_subpage.items():
                   if title not in source_subpage:
                       self.deletePage(page['id'])

    def getSpace(self, space_key, expand='', label=None, status=None):
        return self.get( '/rest/api/space/%s' % space_key, expand=expand, label=label, status=status)

    def listSpaces(self, expand='', status=None, type=None, label=None):
        return self.iterate('get', '/rest/api/space', expand=expand, status=status, type=type, label=label)

    def getPage(self, page_id, expand='', status='current', version=None):
        if isinstance(expand, (list, set)):
            expand=",".join(expand)

        if not page_id.startswith('/rest'):
            if str(page_id).isdigit():
                page_id = '/rest/api/content/%s' % page_id
            else:
                pages = [p for p in self.getPages(page_id, expand=expand)]
                assert len(pages) == 1
                return pages[0]

        return Page(self, self.get( page_id, expand=expand, status=status, version=version), expand=expand)

    def movePage(self, page, parent):
        return self.get('/pages/movepage.action',
            pageId = page.id,
            spaceKey = parent['spacekey'],
            targetTitle = parent['title'],
            position = 'append'
            )

    def getPages(self, cql=None, expand=[], filter=None, state=None, pages=None, version=None):
        """
        state is comala workflow state here
        """
        logger.info("getPages cql=%s, expand=%s, filter=%s, state=%s", cql, expand, filter, state )
        if not expand:
            expand = []

        if pages is not None:
            cql = self.resolveCQL(pages)

        if state is not None:
            _cql = '(%s) and state = "%s"' % (cql, state)
            for page in self.getPages(_cql, expand, filter):
                yield page

            _cql = '(%s) and state != "%s"' % (cql, state)
            for page in self.getPages(_cql):
                result = self.get("/rest/adhocworkflows/1/workflow/%s/states" % page.id)

                logger.debug("result: %s", result)

                for _state in reversed(result['states']):

                    if _state[u'name'] != state:
                        continue

                    version = _state['contentVersion']
                    logger.debug("found state version: %s", version)

                    _expand = expand
                    if filter is not None:
                        _expand.append('body.view')

                    page = self.getPage(page.id, expand=_expand, status='historical', version=version)
                    logger.debug("page: %s", page)
                    for p in self.getPagesWithProperties(page, filter=filter):
                        yield p

                    break

        elif filter is not None:
            for page in self.getPagesWithProperties(cql, filter=filter, expand=expand, version=version):
                yield page

        else:
            for page in self.iterate('findPages', cql=cql, expand=expand):
                yield Page(self, page, expand)

    def getSpaceHomePage(self, space_key):
        logger.info("space_key: %s", space_key)
        homepage = self.getSpace(space_key, expand='homepage')['homepage']['id']
        return homepage

    def convertWikiToStorage(self, content):
        """convert wiki to storage representation

        Returns storage representation.
        """
        return self.post('/rest/api/contentbody/convert/storage',
            value=content, representation='wiki')['value']

    def getLabels(self, page_id):
        return self.get('/rest/api/content/%s/label' % page_id)

    def addLabels(self, page_id, labels):
        if not isinstance(labels, list):
            labels = [ labels ]
        if not isinstance(labels[0], dict):
            labels = [ dict(prefix="global", name=lbl) for lbl in labels ]

        logger.info("labels: %s", labels)
        return self.post('/rest/api/content/%s/label' % page_id, labels)

    def deleteLabels(self, page_id, labels):
        if not isinstance(labels, list):
            labels = [ labels ]
        result = []
        logger.info("remove labels: %s", labels)
        for label in labels:
            result.append(self.delete('/rest/api/content/%s/label/%s' % (page_id, label)))
        return result

    def updatePage(self, id, title, body=None, version=None, type='page', storage=None, wiki=None):
        if not isinstance(version, dict):
            version = {'number': int(version)}

        if storage is not None:
            body = {
                'storage': {
                    'value': storage,
                    'representation': 'storage'
                }
            }
        if wiki is not None:
            body = {
                'storage': {
                    'value': wiki,
                    'representation': 'wiki'
                }
            }

        return self.put('/rest/api/content/%s' % id,
            version = version,
            type    = type,
            title   = title,
            body    = body
        )

    def editPages(self, cql, editor, filter=None):
        """
        Editor works with mustache templates and jQuery assingments.

        editor: must be either a list (of actions) or must be a dictionary with following items:

            * `templates` - a dictionary of templates, which can be used in
              edit actions
            * `partials` - a dictionary of partials, which can be used in
              templates
            * `actions` - an array of editor actions, where each item is a
              dictionary with following items:

              * `select` - (required) a jQuery selector
              * `action` - a jQuery method how to apply content to selection
                (default: html)
              * `content` - content to be applied. If not present, `template`
                and `data` must be present
              * `type` - type of content.  May be either `storage` or `wiki`.
                Default is `storage`
              * `templates` - a local set of templates to be overridden
              * `template` - a template to which data is applied to, for
                generating content.
              * `data` - data to be applied to template
        """

        if not isinstance(editor, StorageEditor):
            editor = StorageEditor(self, **editor)

        for page in self.getPages(cql, filter=filter, expand=['body.storage', 'version']):
            yield page, editor.edit(page)


    def getPageVersion(self, page_id):
        data = self.get('/rest/api/content/%s' % page_id)
        return data['version']['number']

    def findPages(self, pageSpec='', expand='', limit='', start='', cql=''):
        '''pageSpec may be page_id, "space:path" or CQL (see https://developer.atlassian.com/confdev/confluence-server-rest-api/advanced-searching-using-cql)'''

        if pageSpec:
            cql = self.resolveCQL(pageSpec)
        if isinstance(expand, (list, set)):
            expand = ','.join(list(expand))
        return self.get('/rest/api/content/search', cql = cql, expand=expand, limit=limit, start=start)

    def iterate(self, method, *args, **kwargs):
        if 'start' not in kwargs:
            kwargs['start'] = 0
        if 'limit' not in kwargs:
            kwargs['limit'] = -1

        start = kwargs['start']
        limit = 25 # confluence max
        done = False
        maxResults = kwargs['limit']
        if maxResults < 0:
            maxResults = 10000000000

        while True:
            kwargs['start'] = start
            kwargs['limit'] = limit
            result = getattr(self, method)(*args, **kwargs)

            for item in result['results']:
                logger.info("item_id: %s", item['id'])
                yield item
                maxResults -= 1
                if maxResults <= 0:
                    break

            if maxResults <= 0:
                break

            start += limit
            if result['size'] < result['limit']:
                break

            logger.info("next round: start=%s, limit=%s, size=%s, result_limit=%s", start, limit, result['size'], result['limit'])

    SPACE_PAGE_REF = re.compile(r'^([A-Z]*):(.*)$')
    PAGE_REF = re.compile(r'^:(.*)$')
    PAGE_ID = re.compile(r'^(\d+)$')
    PAGE_URI = re.compile(r'api/content/(\d+)$')
    PARENT = re.compile(r'(.*)>$')
    ANCESTOR = re.compile(r'(.*)>>$')

    def resolveCQL(self, ref):
        """resolve some string to CQL query

        :param ref:
            resolve this reverence to a valid CQL

            * ``SPACE:page title`` -> ``space = SPACE and title = "page title"``
            * ``SPACE:page title>`` -> `` parent = <id of specified page>
            * ``SPACE:page title>>`` -> `` ancestor = <id of specified page>
            * ``:page title`` ->  ``title = "page title"``
            * ``12345`` -> ``ID = 12345``
            * ends with ``api/content/12345`` -> ``ID = 12345``
            * else assume ``ref`` is already CQL

        :return:
            CQL
        """
        def match(RE):
            m = RE.search(ref)
            if m:
                self.mob = m.groups()
                return True
            else:
                return False

        if not is_string(ref):
            ref = str(ref)

        ref = ref.strip()

        if match(self.ANCESTOR):
            query = self.resolveCQL(*self.mob)
            p = self.getPage(query)
            return u"ancestor = {}".format(p.id)

        if match(self.PARENT):
            query = self.resolveCQL(*self.mob)
            p = self.getPage(query)
            return u"parent = {}".format(p.id)

        if match(self.SPACE_PAGE_REF):
            return u"space = {} AND title  = \"{}\"".format(*self.mob)

        if match(self.PAGE_REF):
            return u"title  = \"{}\"".format(*self.mob)

        if match(self.PAGE_ID):
            return u"ID  = {}".format(*self.mob)

        if match(self.PAGE_URI):
            return u"ID  = {}".format(*self.mob)

        return ref

    def cwInfo(self, page, expand=[]):
        """return comala workflow information about current page"""

        if isinstance(expand, list):
            expand = ','.join(expand)
        return self.get('/rest/cw/1/content/%s' % page.id, expand=expand)

    def cwApprove(self, page, name, note=None):
        """approve a page"""
        return self.post(
            '/rest/adhocworkflows/latest/approval/%s/approve' % page.id,
            name = name,
            note = note)

    def cwReject(self, page, name, note=None):
        """approve a page"""
        return self.post(
            '/rest/adhocworkflows/latest/approval/%s/reject' % page.id,
            name = name,
            note = note)


    def setPageProperties(self, document):
        pages = document.pop('pages', None)

        if pages is not None:
            for page in pages:
                _doc = document.copy()
                _doc.update(page)
                for p in self.setPageProperties(_doc):
                    yield p

        cql = None
        if 'page' in document:
            cql = self.resolveCQL(document['page'])
        elif 'cql' in document:
            cql = document['cql']

        if cql is not None:
            editor = PagePropertiesEditor(confluence=self, **document)

            found = False
            for page in self.getPagesWithProperties(cql, expand=['body.storage', 'version']):
                new_content = editor.edit(page)
                found = True
                yield dict(
                    page    = page,
                    content = new_content,
                    result  = self.updatePage(
                        id = page['id'],
                        version = int(page['version']['number'])+1,
                        title   = page['title'],
                        storage = new_content
                    ))

            if not found:
                new_content = editor.edit()
                (space, title) = document['page'].split(':', 1)
                yield dict(
                    page    = dict(
                        spacekey = space,
                        title    = title,
                        ),
                    content = new_content,
                    result  = self.createPage(
                        space = space,
                        title = title,
                        storage = new_content,
                        parent = document.get('parent', None)
                    ))



    PAGE_PROP_FILTER = re.compile(r'^(?:(.*?)([!=])=(.*)|!(.*)|(.*)\?)$')
    def getPagesWithProperties(self, cql, filter=None, expand=[], state=None, **options):
        """Either pass CQL or a page having body.view expanded.
        """
        page_prop_filters = []

        logger.info("cql: '%s', filter: %s", cql, filter)

        if isinstance(cql, Page):
            pages = [ cql ]
        else:
            cql = self.resolveCQL(cql)
            pages = self.getPages(cql, state=state, expand=expand + ['body.view'])

        if filter is not None:
            if not isinstance(filter, list):
                filter = [ filter ]

            for item in filter:
                if is_string(item):
                    m = self.PAGE_PROP_FILTER.search(item)
                    if m:
                        (name, cmp, value, not_exists, present) = m.groups()
                        page_prop_filters.append(dict(name=name, cmp=cmp, value=value, not_exists=not_exists, present=present))
                else:
                    page_prop_filters.append(item)

        def page_prop_filterer(page):
            if not len(page_prop_filters):
                return True

            result = True
            for f in page_prop_filters:
                value = page.getPageProperty(f['name'])
                cmp = f['cmp']
                not_exists = f['not_exists']
                present = f['present']

                logger.info("filter: %s, value = %s", f, value)

                if cmp == '=':
                    if isinstance(value, list):
                        result = result and f['value'] in value
                    else:
                        result = result and value == f['value']
                elif not_exists:
                    value = page.getPageProperty(f['not_exists'])
                    result = result and value is None

                elif present:
                    value = page.getPageProperty(f['present'])
                    result = result and value is not None

                else:
                    if isinstance(value, list):
                        result = result and f['value'] not in value
                    else:
                        result = result and value != f['value']

            logger.info("filterer result = %s", result)
            return result


        for page in filter_func(page_prop_filterer, pages):
            yield page

    def getContentId(self, page):
        (space, title, id) = self.extractPage(page)
        return id

    def extractPage(self, pageSpec):
        results = self.findPages(pageSpec, expand='space')

        if results is None:
            return None

        assert results['size'] == 1, "Ambigious search: %s" % pageSpec

        page = results['results'][0]
        return page['space']['key'], page['title'], page['id']

    def createPage(self, space, title, storage, parent=None):
        data = dict(
            title = title,
            type  = 'page',
            space = {'key': space},
            body  = {'storage': {'value': storage, 'representation': 'storage'}}
        )

        if parent is not None:
            data['ancestors'] = [{'id': self.getContentId(parent)}]

        return self.post('/rest/api/content', **data)

    def getChildren(self, page_id, type=None, expand=''):
        if not page_id.startswith('/rest'):
            page_id = '/rest/api/content/%s' % page_id
        if type is None:
            url = page_id +"/child"
        else:
            url = page_id + "/child/"+type
        return self.get(url, expand=expand)['results']



import yaml

def get_list_data(data):
    #print "data", data
    if isinstance(data, basestring):
        ac = []
        for doc in yaml.safe_load_all(data):
            if isinstance(doc, list):
                ac += doc
            else:
                ac.append(doc)
        data = ac
    if not isinstance(data, list):
        data = [ data ]
    #print "data", data
    return data

import pyaml, yaml

from pyaml import UnsafePrettyYAMLDumper

def represent_stringish(dumper, data):
    data = unicode(data) # read the comment above

    style = dumper.pyaml_string_val_style
    if not style:
        style = 'plain'

        if '\n' not in data and ('@' in data or '{' in data or '}' in data):
            style = "'"

        elif '\n' in data or not data or data == '-' or data[0] in '!&*[?':
            if 0:
                style = 'literal'
                if '\n' in data[:-1]:
                    for line in data.splitlines():
                        if len(line) > dumper.best_width: break
                else: style = '|'

            style = "|"

    return yaml.representer.ScalarNode('tag:yaml.org,2002:str', data, style=style)

for str_type in {bytes, unicode}:
    UnsafePrettyYAMLDumper.add_representer(
        str_type, represent_stringish )



from .cli import *

space_command = command.add_subcommands('space', help="working spaces")

@space_command('ls', arg_format(default="{key:10s} {name}"), arg_write_format(default='format'), arg_expand, arg_status, arg_label,
    arg('-t', '--type', choices=('global', 'personal', 'all'), default="global", help="space type"),
    arg_field,
    )
def space_list(config):
    # type, status, label, expand
    kwargs = config.dict('expand', 'status', 'type', 'label')
    if kwargs['type'] == 'all':
        del kwargs['type']

    results = [ r for r in config.confluence_api.listSpaces(**kwargs) ]

    if config.get('write') == 'format':
        if config.get('format'):
            for result in results:
                if '{}' in config['format']:
                    fields = [ result[f] for f in config['field'] ]
                    print(config['format'].format(*fields))
                else:
                    print(unicode(config['format']).format(**result))
    else:
        if config['write'] == 'json':
            import json
            json.dump(results, sys.stdout)

        elif config['write'] == 'yaml':
            import pyaml
            pyaml.p(results)



from .cli import command, arg, optarg_cql, arg_filter, arg_parent, arg_cql, arg_message, arg_expand
import pyaml

cw_command = command.add_subcommands('cw', help="comala workflows")

def print_info(page, result):
    pyaml.p(dict(
        id = page.id,
        spacekey = page.spacekey,
        title = page.title,
        cwInfo = result))

@cw_command('info',
    arg_cql,
    arg_expand
)
def cw_info(config):
    """get workflow information about given pages
    """
    confluence = config.getConfluenceAPI()
    first = True
    for page in confluence.getPages(pages=config.cql):
        if not first:
            print "---"
        result = confluence.cwInfo(page, config.expand)
        first = False
        print_info(page, result)

@cw_command('approve', arg_cql, arg_message,
    arg('-n', '--name', help="approval name")
)
def cw_approve(config):
    """approve a page
    """
    confluence = config.getConfluenceAPI()

    first = True
    for page in confluence.getPages(pages=config.cql):
        if not first:
            print "---"
        first = False
        if not config.name:
            result = confluence.cwInfo(page,expand='approvals')

            if 'approvals' not in result:
                result['message'] = 'approvals not in result'
                print_info(page, result)
                continue

            if len(result['approvals']) > 1:
                names = ", ".join([ a['name'] for a in result['approvals'] ])
                raise RuntimeError("please pass --name with one of %s" % (names,))

            if result['state']['final']:
                print_info(page, result)
                continue

            if len(result['approvals']) != 1:
                result['message'] = 'ambigious approvals, specify one with -n'
                print_info(page, result)
                continue

            name = result['approvals'][0]['name']

        else:
            name = config.name

        result = confluence.cwApprove(page, name=name, note=config.get('message', ''))
        print_info(page, result)


@cw_command('reject', arg_cql, arg_message,
    arg('-n', '--name', help="approval name"),
)
def cw_reject(config):
    """reject a page
    """
    confluence = config.getConfluenceAPI()

    first = True
    for page in confluence.getPages(pages=config.cql):
        if not first:
            print "---"
        first = False
        if not config.name:
            result = confluence.cwInfo(page,expand='approvals')
            if len(result['approvals']) > 1:
                names = ", ".join([ a['name'] for a in result['approvals'] ])
                raise RuntimeError("please pass --name with one of %s" % (names,))

            if result['state']['final']:
                print_info(page, result)
                continue

            if len(result['approvals']) != 1:
                result['message'] = 'ambigious approvals, specify one with -n'
                print_info(page, result)
                continue

            name = result['approvals']['name']

        else:
            name = config.name

        result = confluence.cwReject(page, name=name, note=config.get('message', ''))

        print_info(page, result)



from pyquery import PyQuery
from .storage_editor import storage_query

class DataGenerator:
    def __init__(self, generator, api=None):
        self.api = api
        self.generator = generator
        print generator

    def __call__(self, data):
        if "<ri:" in data or '<ac:' in data:
            self.query = storage_query
        else:
            self.query = PyQuery

        return self.generate(self.query(data)(":root"), **self.generator)

    def get_value(self, element, data, method, *args):
        selection = self.get_selection(element, data)
        return getattr(selection, method)(*args)

    def get_selection(self, element, data):
        for method in ['find', 'children', 'parents', 'closest', 'filter', 'next_all', 'prev_all', 'siblings']:
            if method in data:
                if not data[method]:
                    return getattr(element, method)()
                else:
                    return getattr(element, method)(data[method])
        if 'eq' in data:
            return element.eq(int(data['eq']))

        if 'toplevel' in data:
            for e in element.find(data['toplevel']).items():
                for item in e.parents():
                    if item is
                # only those elements, which have no data['toplevel'] parents,
                # which are under element
                if e
            for element in

        return element

    def meet_conditions(self, element, data):
        result = True
        if 'is' in data:
            result = result and element.is_(data['is'])
        if 'isnt' in data:
            result = result and element.not_(data['isnt'])
        if 'not' in data:
            result = result and element.not_(data['not'])
        if 'has' in data:
            if isinstance(data['has'], dict):
                for k,v in data['has'].items():
                    result = result and len(getattr(element, k)(v)) > 0
            else:
                result = result and len(element.parents(data['has'])) > 0

        return result

    def generate(self, element, **data):
        if 'do' in data:
            selection = self.get_selection(element, data)
            if isinstance(data['do'], list):
                for item in data['do']:
                    if self.meet_conditions(selection, item):
                        return self.generate(e, **item)
            else:
                return self.generate(selection, **data)

        if 'list' in data:
            result = []

            for e in self.get_selection(element, data):
                e = self.query(e)

                if data['list'] == 'text':
                    result.append(e.text())
                elif data['list'] == 'html':
                    result.append(e.html())
                else:
                    for item in data['list']:
                        if not self.meet_conditions(element, item): continue
                        result.append(self.generate(e, **item))
            return result

        if 'object' in data:
            result = {}
            for e in self.get_selection(element, data):
                e = self.query(e)

                for item in data['object']:
                    if not self.meet_conditions(element, item): continue
                    result[item['name']] = self.generate(e, **item)

            return result

        if 'attr' in data:
            return self.get_value(element, data, 'attr', data['attr'])

        if data.get('return'):
            return self.get_value(element, data, data['return'])

        else:
            return self.get_value(element, data, 'text')



import pyquery, sys, re
from pyquery.pyquery import fromstring, no_default

PY3k = sys.version_info >= (3,)

if PY3k:
    from urllib.parse import urlencode
    from urllib.parse import urljoin
    basestring = (str, bytes)
    unicode = str
else:
    from urllib import urlencode  # NOQA
    from urlparse import urljoin  # NOQA


class MyQuery(pyquery.PyQuery):

    def __init__(self, *args, **kwargs):
        namespaces = kwargs.get('namespaces')
#        if namespaces:
        if 1:
            kwargs['parser'] = 'xml'

            length = len(args)
            if length == 1:
                selector, context = None, args[0]
            elif length == 2:
                selector, context = args

            if isinstance(context, basestring):
                root = fromstring(
                    self._wrap_root(context, namespaces),
                    kwargs.get('parser')
                    )[0]

                if selector is not None:
                    args = (selector, root.getchildren())
                else:
                    args = (root.getchildren(),)

        super(MyQuery,self).__init__(*args, **kwargs)


    def _wrap_root(self, context, namespaces=None):
        if namespaces is None:
            if hasattr(self, 'namespaces'):
                namespaces = self.namespaces
            else:
                namespaces = {}

        attrs = " ".join([ 'xmlns:%s="%s"' % (k,v) for (k,v) in namespaces.items() ])

        result = unicode("<root %s>" % attrs)+context+unicode('</root>')
        #print result

        return result


    def _get_root(self, value):
        """In case of creating the value from string, namespace has to be passed
        """
        if isinstance(value, basestring):
            root = fromstring(self._wrap_root(value), self.parser)[0]
            # TODO: warn if root has more than one child?
            value = root.getchildren()[0]

        return super(MyQuery, self)._get_root(value)


    def _copy(self, *args, **kwargs):
        """must also set parser"""
        kwargs.setdefault('parser', self.parser)
        return super(MyQuery,self)._copy(*args,**kwargs)

    HTML_TAG = re.compile(r'(<!--.*?-->|<[^>]*>)', re.DOTALL)
    XMLNS = re.compile(r'\sxmlns:[\w\-]+="[^"]*"')
    def strip_namespaces(self, html):
        if not isinstance(html, basestring):
            return html

        html_items = self.HTML_TAG.split(html)
        #print "html_items: %s" % html_items
        for i,part in enumerate(html_items):
#            print part
            if part.startswith('<'):
                if 'xmlns:' in part:
                    html_items[i] = x = self.XMLNS.sub('', part)

        return html.__class__('').join(html_items)

    def __unicode__(self):
        return self.strip_namespaces(super(MyQuery,self).__unicode__())

    def __str__(self):
        return self.strip_namespaces(super(MyQuery,self).__str__())

    def html(self, value=no_default, **kwargs):
        """This cannot be wrapped and needs (almost) full override.
        """
        if value is no_default:
            html = super(MyQuery,self).html(value, **kwargs)
            return self.strip_namespaces(html)

        else:
            if isinstance(value, self.__class__):
                new_html = unicode(value)
            elif isinstance(value, basestring):
                new_html = value
            elif not value:
                new_html = ''
            else:
                raise ValueError(type(value))

            for tag in self:
                for child in tag.getchildren():
                    tag.remove(child)
                root = fromstring(
                    self._wrap_root(new_html),
                    self.parser)[0]
                children = root.getchildren()
                if children:
                    tag.extend(children)
                tag.text = root.text
                tag.tail = root.tail

        return self



import yaml, pyaml, sys
from difflib import Differ
from .cli import command, arg, optarg_cql, arg_filter, arg_parent, arg_add_label
from ..storage_editor import StorageEditor

# @command('create', arg_parent, arg_label, arg_space, arg("pagespec")
# )
# def cmd_create(config):


@command('edit',
    optarg_cql,
    arg_filter,
    arg_parent,
    arg_add_label,
    arg('file', nargs="?", help="file to read data from"),
    # need arg_group
    arg('--show', action="store_true", help="show new content"),
    arg('--diff', action="store_true", help="show diff"),
    )
def cmd_edit(config):
    """edit a confluence page using CSS selections

    Pass a dictionary in YAML or JSON format via STDIN or file to
    confluence-tool, which defines edit actions to edit all matching pages.

    """

    confluence = config.getConfluenceAPI()
    first = True

    if not config['file']:
        editor_config = yaml.safe_load(sys.stdin)
    else:
        with open(fn, 'r') as f:
            editor_config = yaml.safe_load(f)

    if 'page' in editor_config:
        cql = confluence.resolveCQL(editor_config.pop('page'))
    if 'cql' in editor_config:
        cql = editor_config.pop('cql')
    if config['cql']:
        cql = config['cql']

    #editor = StorageEditor(confluence, **editor_config)

    for page,content in confluence.editPages(cql, filter=config.filter, editor=editor_config):
        if not first:
            print "---"
        first = False

        from html5print import HTMLBeautifier
        b = HTMLBeautifier.beautify

        if config.show:
            p = page.dict('id', 'spacekey', 'title')

            p['content'] = b(content, 2)
            pyaml.p(p)

        elif config.diff:
            p = page.dict('id', 'spacekey', 'title')

            old = b(page['body']['storage']['value']).splitlines(1)
            new = b(content).splitlines(1)

            d = Differ()
            result = list(d.compare(old, new))

            p['diff'] = ''.join(result)
            pyaml.p(p)

        else:
            p = page.dict('id', 'title', 'version')
            p['storage'] = content
            p['version'] = int(page['version']['number'])+1

            result = confluence.updatePage(**p)
            pyaml.p(result)


@command('move', optarg_cql, arg_filter, arg_parent)
def move(config):
    confluence = config.getConfluenceAPI()
    first = True

    parent = confluence.getPage(confluence.resolveCQL(config.parent))
    cql = confluence.resolveCQL(config.cql)
    filter = config.filter

    for page in confluence.getPages(cql, filter=filter, expand=['version']):
        result = confluence.movePage(page, parent=parent)

        if not first:
            print "---"

        pyaml.p(result)

        first = False

    if first:
        print "could not find a page matching %s (%s)" % (cql, filter)



import sys, re
from .cli import command, arg, arg_cql, arg_filter
import pyaml
from ..confluence_api import ConfluenceError

@command('labels',
    arg_cql,
    arg_filter,
    arg('-a', '--add', action="append", help="label to add"),
    arg('-r', '--remove', action="append", help="label to remove"),
    arg('-q', '--quiet', action="store_true", help="do not show labels of the page"),
    )
def cmd_page_prop_get(config):
    confluence = config.getConfluenceAPI()
    results = []
    config['cql'] = confluence.resolveCQL(config['cql'])
    first = True
    for pp in confluence.getPages(**config.dict('cql', 'filter')):
        if first:
            first = False
        else:
            print "---"

        if config.get('add'):
            results.append(confluence.addLabels(pp.id, config['add']))
        if config.get('remove'):
            try:
                results.append(confluence.deleteLabels(pp.id, config['remove']))
            except ConfluenceError as e:
                print "Warning: %s" % e
                continue

        pyaml.p(pp.dict('id', 'spacekey', 'title', 'labels'))



from os.path import expanduser
from confluence_tool import ConfluenceError, ConfluenceAPI
import pyaml, yaml
import logging

logging.basicConfig()

import keyring

class ConfigFileMissing(StandardError):
    pass

class Config:
    def __init__(self, **args):
        self.args = args

    def __getattr__(self, name):
        if name == 'config':
            self.config = self.readConfig()
            return self.config

        if name == 'config_file':
            if self.args.get('config_file'):
                self.config_file = self.args['config_file']
            else:
                self.config_file = expanduser('~/.confluence-tool.yaml')

            return self.config_file

        if name == 'confluence_api':
            self.confluence_api = self.getConfluenceAPI()
            return self.confluence_api

        return self.get(name)

    def dict(self, *args):
        result = {}
        for k in args:
            result[k] = self.get(k)
        return result

    def __getitem__(self, name):
        return self.args[name]

    def __setitem__(self, name, value):
        self.args[name] = value

    def __contains__(self, name):
        return name in self.args

    def get(self, name, default=None):
        try:
            return self[name]
        except KeyError:
            return default

    def readConfig(self):
        try:
            with open(self.config_file, 'r') as f:
                return yaml.load(f)
        except Exception as e:
            if self.args.get('debug'):
                import traceback
                traceback.print_exc()

            raise ConfigFileMissing()

    def writeConfig(self):
        with open(self.config_file, 'w') as f:
            pyaml.p(self.config, file=f)

    def getConfig(self):
        result = {}

        result.update(
            baseurl  = self.args.get('baseurl'),
            username = self.args.get('username'),
            password = self.args.get('password'),
            )

        if not result['baseurl'] or self.args.get('config'):
            config_name = self.args.get('config', 'default') or 'default'
            result.update(**self.config[config_name])

        if result['username'] and not result['password']:
            baseurl = result['baseurl']
            password = keyring.get_password('confluence-tool '+baseurl, result['username'])
            result['password'] = password

        return result

    def setConfig(self, update_password=False):
        config_name = self.args.get('config', 'default')

        try:
            config = self.config or {}
        except ConfigFileMissing:
            config = {}

        old_baseurl = config[config_name].get('baseurl')
        old_username = config[config_name].get('username')

        print "ou: %s" % old_username

        baseurl  = self.args.get('baseurl')
        username = self.args.get('username')

        if baseurl is None:
            baseurl = old_baseurl
        if username is None:
            username = old_username

        if config_name in config:
            old_baseurl = config[config_name].get('baseurl')
            old_username = config[config_name].get('username')
            if old_baseurl != baseurl or old_username != username:
                try:
                    keyring.delete_password("confluence-tool "+old_baseurl, old_username)
                except:
                    pass

        config[config_name] = cfg = dict(
            baseurl  = baseurl,
            username = username,
        )

        print "Password will be stored in your systems keyring."
        import getpass
        password = getpass.getpass("Password (%s): " % username)
        keyring.set_password('confluence-tool '+baseurl, username, password)

        for k,v in cfg.items():
            if v is None:
                del cfg[k]

        self.config = config

        self.writeConfig()

    def rmConfig(self):
        config_name = self.args.get('config', 'default')

        try:
            config = self.config or {}
        except ConfigFileMissing:
            config = {}

        if config_name not in config:
            return

        keyring.delete_password('confluence-tool '+baseurl, username)

        del config[config_name]
        self.writeConfig()


    def getConfluenceAPI(self):
        return ConfluenceAPI(self.getConfig())

from .cli import command, arg

confluence_tool_config = {}

def main(argv=None):

    def config_factory(args, **kwargs):
        global confluence_tool_config
        confluence_tool_config = Config(**vars(args))
        if args.debug:
            log = logging.getLogger()
            log.setLevel(logging.DEBUG)

        return [confluence_tool_config], {}

    try:
        return command.execute(argv, compile=config_factory)

    except ConfigFileMissing as e:
        if confluence_tool_config.get('debug'):
            import traceback
            traceback.print_exc()
            return 1

        #print "Config file missing.  Please run 'confluence-tool config' or specify --baseurl"
        return 1

    except StandardError as e:
        if confluence_tool_config.get('debug'):
            import traceback
            traceback.print_exc()
            return 1
        else:
            print (u"%s" % e).encode('utf-8')
            return 1

import edit
import page_prop
import show
import config
import labels
import comala_workflow
import space



"""
Confluence Commandline Interface

test

x

y

"""

import pyaml, yaml, sys, json, logging
import argparse
from os.path import expanduser
from confluence_tool import ConfluenceError, ConfluenceAPI
from textwrap import dedent
import difflib

def have_data_on_stdin():
    import sys
    import select

    if select.select([sys.stdin,],[],[],0.0)[0]:
        print "Have data!"
    else:
        print "No data"

from pprint import pprint

logging.basicConfig()

OUTPUT_FORMAT = 'yaml'

class Config:
    def __init__(self, **args):
        self.args = args

    def __getattr__(self, name):
        if name == 'config':
            self.config = self.readConfig()
            return self.config

        if name == 'config_file':
            if self.args['config_file']:
                self.config_file = self.args['config_file']
            else:
                self.config_file = expanduser('~/.confluence-tool.yaml')

            return self.config_file

        if name == 'confluence_api':
            self.confluence_api = self.getConfluenceAPI()
            return self.confluence_api

        try:
            return self[name]
        except KeyError:
            pass

        raise AttributeError(name)


    def dict(self, *args):
        result = {}
        for k in args:
            result[k] = self[k]
        return result

    def __getitem__(self, name):
        return self.args[name]

    def readConfig(self):
        with open(self.config_file, 'r') as f:
            return yaml.load(f)

    def writeConfig(self):
        with open(self.config_file, 'w') as f:
            pyaml.p(self.config, file=f)

    def getConfig(self):
        result = {}

        result.update(
            baseurl  = self.args.get('baseurl'),
            username = self.args.get('username'),
            password = self.args.get('password'),
            )

        if not result['baseurl'] or self.args.get('config'):
            config_name = self.args.get('config', 'default') or 'default'
            result.update(**self.config[config_name])

        return result

    def setConfig(self):
        config_name = self.args.get('config', 'default')
        self.config[config_name] = dict(
            baseurl  = self.args.get('baseurl'),
            username = self.args.get('username'),
            password = self.args.get('password'),
        )
        self.writeConfig()

    def getConfluenceAPI(self):
        return ConfluenceAPI(self.getConfig())



def dump(obj):
    if OUTPUT_FORMAT == 'yaml':
        try:
            pyaml.p(obj)
            return
        except ImportError:
            pass

    json.dump(obj, sys.stdout, indent=2)



def main(argv):
    args = argparser.parse_args(argv)
    opts = vars(args).copy()
    del opts['action']

    global OUTPUT_FORMAT
    if 'json' in opts:
        if opts['json']:
            OUTPUT_FORMAT = 'json'
        del opts['json']

    config = Config(**opts)
    try:
        args.action(config)
    except ConfluenceError as e:
        dump(json.loads(str(e)))
    except Exception as e:
        if config.debug:
            import traceback
            traceback.print_exc()
        else:
            sys.stderr.write("%s\n" % e)

        return 1


if __name__ == '__main__':

    sys.exit(main(sys.argv[1:]) or 0)



from .cli import *
import sys

@command('show',
    positional_arg_cql,
    arg_filter,
    arg_expand,
    arg_write_format,
    arg_format,
    arg_state,
    mutually_exclusive(
        arg('--storage', action="store_true", help="convenience for: -e 'body.storage' -F '{body[storage][value]}'"),
        arg('--html', action="store_true", help="convenience for: -e 'body.view' -F '{body[view][value]}'"),
        arg('--ls', action="store_true", help="convenience for: -F '{id} {spacekey} {title}'"),
    ),
    arg('-B', '--beautify', action="store_true", help="beautify body.storage.value or body.view.value, if present"),
#    arg('-d', '--data', help="filename containing data selector in YAML or JSON format"),
    arg('field', nargs="*", help='field to dump')
)
def show(config):
    """show a confluence item


    """


    """show a confluence item

    If specifying data selector file, there is added a special field
    body['data'] to each page.

    Format of selector file:

    select: <CSS Selector for jQuery>
    list:
        - attr: href
          name: href

    Following keys:

    - list: each item produces a list item.  Input is current selected element.
    - object: each item produces a object property.  Each item must have "name"
        key for specifing the properties name
    - attr: read named attribute from current selected element
    - select: find elements using this selector.  For each of them apply current
      spec
    - text: 'text' or 'html'.  this is the default and default is to return
        'text'.

    Examples:

     - `ct show "SpaceName:Some Title"`

        gets page information which can be expanded

     - `ct show "SpaceName:Some Title" -e body`

        gets page information with expanded body (can also be expanded)

     - `ct show "SpaceName:Some Title" -e body,container`

        gets page information and expand more than one

     - `ct show "SpaceName:Some Title" -e 'body.view' -F '{body[view][value]}'`

        gets page information with more detailed expanded body (this is --html)

     - `ct show "SpaceName:Some Title" -F {_links[self]}`

        gets "self"-link expanded from "_links" from "Some Title"
    """
    first = True

    mustache, format, printf = None,None,None

    output_filter = lambda x: x

    if not config.get('format') and not config.get('expand'):
        if config.get('html'):
            config['format'] = u'{body[view][value]}'
            config['expand'] = 'body.view'

            from html5print import HTMLBeautifier
            output_filter = lambda x: HTMLBeautifier.beautify(x, 4)

        elif config.get('storage'):
            config['format'] = u'{body[storage][value]}'
            config['expand'] = 'body.storage'

            from html5print import HTMLBeautifier
            output_filter = lambda x: HTMLBeautifier.beautify(x, 4)

        elif config.get('ls'):
            config['format'] = u'{id}  {spacekey}  {title}'
            config['field'] = ['id', 'spacekey', 'title']

    results = []
    kwargs = config.dict('cql', 'expand', 'filter', 'state')
    kwargs['cql'] = config.confluence_api.resolveCQL(kwargs['cql'])

    for page in config.confluence_api.getPages(**kwargs):
        rec = page.dict(*config['field'])
        if config.get('beautify'):
            from html5print import HTMLBeautifier
            if rec.get('body', {}).get('storage', {}).get('value'):
                rec['body']['storage']['value'] = HTMLBeautifier.beautify(rec['body']['storage']['value'], 4)
            if rec.get('body', {}).get('view', {}).get('value'):
                rec['body']['view']['value'] = HTMLBeautifier.beautify(rec['body']['view']['value'], 4)

        results.append(rec)

    if config.get('format'):
        for result in results:
            if '{}' in config['format']:
                fields = [ result[f] for f in config['field'] ]
                print config['format'].format(*fields)

            print output_filter(unicode(config['format']).format(**result))

    elif config.get('data'):
        if config.get('data') == '-':
            data = get_list_data(sys.stdin.read())
        else:
            with open(config.get('data'), 'r') as f:
                data = get_list_data(f.read())

        from ..data_generator import generate_data

        pyaml.p(generate_data(data, results))

    else:
        if len(results) == 1:
            results = results[0]

        if config['write'] == 'json':
            import json
            json.dump(results, sys.stdout)

        elif config['write'] == 'yaml':
            import pyaml
            pyaml.p(results)



from HTMLParser import HTMLParser
htmlparser = HTMLParser()

from lxml.etree import XMLSyntaxError


import logging
log = logging.getLogger('confluence-tool.page')

class Page(object):
    def __init__(self, api, data, expand=None):
        self.api = api
        self.data = data

        if 'body' in self.data:
            body = self.data['body']

            if 'storage' in body:
                body = body['storage']

                log.debug("body: %s", body['value'])
                body['value'] = htmlparser.unescape(body['value'])
                log.debug("unescaped body: %s", body['value'])

            elif 'view' in body:
                body = body['view']
                log.debug("unescaped body")
                body['value'] = htmlparser.unescape(body['value'])


        if hasattr(expand, 'split'):
            self.expand = set([ s.strip() for s in expand.split(',') ])
        else:
            self.expand = set(expand)

    def update(self):
        self.api.getPage(self.data['id'], expand=self.expand)

    def __getattr__(self, name):
        if name == 'pageProperty':
            self.pageProperty = self.data['pageProperties'] = dict(self.loadPageProperties())
            return self.pageProperty

        if name == 'labels':
            result = self.api.getLabels(self.data['id'])
            self.labels = [ l['name'] for l in result['results']]
            return self.labels

        if name == 'spacekey':
            return self['spacekey']

        if name in self.data:
            return self.data[name]
        raise AttributeError(name)

    def __getitem__(self, name):
        if name == 'pageProperties':
            return self.pageProperty
        if name == 'labels':
            return self.labels

        if name == 'spacekey':
            return self.data['_expandable']['space'].split("/")[-1]
        else:
            return self.data[name]

        raise KeyError(name)

    def get(self, name, default=None):
        try:
            return self[name]
        except KeyError:
            return default

    def dict(self, *keys):
        '''compose new dictionary from given keys'''
        if not len(keys):
            return self.data

        result = {}
        for key in keys:
            result[key] = self[key]

        return result

    def getPageProperty(self, name, default=None):
        return self.pageProperty.get(name, default)

    def getPageProperties(self, *names):
        for k,v in self.pageProperty.items():
            if len(names):
                if k in names:
                    yield (k,v)
            else:
                yield (k,v)

    def loadPageProperties(self, need_html=False, need_data=False, properties=None, **opts):
        from .page_properties import get_page_properties
        if 'body.view' not in self.expand:
            self.expand.add('body.view')
            self.update()
        html = self.data['body']['view']['value']
        return get_page_properties(html, need_html=need_html, need_data=need_data, properties=properties)
        #self.data['pageProperties'] = dict(get_page_properties(self.data['body']['view']['value']))



import sys, re
from .cli import command, arg, arg_format, optarg_cql, arg_filter, arg_state
import pyaml, yaml

@command('page-prop-get', optarg_cql, arg_filter, arg_format, arg_state,
    arg('--dict', action="store_true", help="transform page properties to dict (key page_id) before output"),
    arg('props', nargs="*", help="properties to retrieve"),
    )
def cmd_page_prop_get(config):
    """\
    Get page properties.

    For each page there is printed a YAML document with `id`, `title` and
    `spacekey`.  Page properties are printed under `pageProperties`.

    Optionally you can pass some page property filter expressions to filter
    pages on page properties additionally to CQL query.

    Please note, when using `--format`:

    For convenience in format string, you can directly refer to page properties
    in format string.  Items referring to page can be fetched with `page_id`,
    `page_title` and `page_spacekey`.

    Examples:

     - `ct page-prop-get "SpaceName:Some Title"`

        gets all page properties from "Some title" in space "Space Name"

     - `ct page-prop-get "SpaceName:Some Title" Pageproperty1 Pageproperty2`

       get Pageproperty1 and Pageproperty2 from "Some title" in space "Space Name"

     - `ct page-prop-get "label = 'some-label'"`

        gets all page properties for all pages with label 'some-label'
    """
    confluence = config.getConfluenceAPI()
    first = True

    if config.get('dict'):
        results = {}

    for pp in confluence.getPagesWithProperties(**config.dict('cql', 'filter', 'state')):

        if config.get('format'):
            try:

                if '{}' in config['format']:
                    fields = [ pp.pageProperty.get(f, '') for f in config['props'] ]
                    print unicode(config['format']).format(*fields).encode('utf-8')
                else:
                    _props = dict(pp.getPageProperties())
                    _props.update(dict([ (k.lower(), v) for (k,v) in _props.items()]))
                    _props.update(page_id=pp.id, page_title=pp.title, page_spacekey=pp.spacekey)

                    print unicode(config['format']).format(**_props).encode('utf-8')
            except UnicodeEncodeError:
                import traceback
                sys.stderr.write("Error formatting %s:%s\n %s\n" % (pp.spacekey, pp.title, repr(dict(pp.getPageProperties()))))
        else:
            result = pp.dict("id", "spacekey", "title")
            result['pageProperties'] = dict(pp.getPageProperties(*config.props))

            if config.get('dict'):
                results[result['id']] = result
            else:
                if not first:
                    print("---")
                else:
                    first = False

                result = pp.dict("id", "spacekey", "title")
                result['pageProperties'] = dict(pp.getPageProperties(*config.props))

                pyaml.p(result)

    if config.get('dict'):
        pyaml.p(results)


@command('page-prop-set',
    arg_filter,
    arg('-p', '--parent', help="specify parent for a page, which might be created"),
    arg('-l', '--label', action="append", help="add these labels to the page"),
    optarg_cql,
    arg('propset', nargs="*", help="property setting expression"),
    arg('file', nargs="*", help="file to read data from")
    )
def cmd_page_prop_set(config):
    """\
    Set page properties.

    # Details

    There are multiple ways of setting page properties.  CQL and page property
    filters are working like in `page-prop-get` command and select pages to
    edit properties.

    # Setting page properties via arguments

    You can add multiple `propset` epressions.  A Propset expression is:

    * `propname:=value` -
      Replace current property with this value.  Value may be JSON or a string.

    * `propname+=value` -
      add value to current property value.

    * `propname-=value` -
      remove value from current property value.

    * `propname--` -
      remove property

    This way default templates for rendering properties are used.


    # Setting page properties via YAML from STDIN

    A document may have following values:

    - `page` - specify a page to be changed
    - `parent` - a page to be parent of a new page
    - `templates` - dictionary of templates with following names:
      - `user` - to render user names.  Gets userkey
      - `page` - to render page Gets spacekey, title
      - `link` - to render link Gets href, caption
      - `list` - to render a list
      - `PROPKEY-TYPE`, where PROPKEY is valid propkey and TYPE is one of the
        above.  This will be used as templates only for that key
    - `pages` - list of documents like this
    - `pagePropertiesEditor` - Define how to change page properties.  This is a
      dictionary with PROPKEY and values define actions.

      An action can be one of the following:
        - `delete` - delete the prop
        - dictionary with one or more of following key/value pairs:
            - `replace` - replace prop's value with given value
            - `add` - add given value/s to prop
            - `remove` - remove given value/s from prop if present
    - `pagePropertiesOrder` - (or short `order`) Define order of page
      properties, in case they are not yet part of document.

    Instead of `pagePropertiesEditor`, you can also specify pageProperties.
    This may be either a dictionary specifying props and values or it may be
    a list of such dictionaries, directly specifying `pagePropertiesOrder`.

    Examples:

     - `ct page-prop-set -p "SpaceName:Parent Title" "SpaceName:Some Title"
       PageProp1:='Content1' 'Page Prop2:=Content 2' 'Page-Prop3:=Content3'`

       Replaces page properties for Page "SpaceName:Some Title" (creates page and
       page-properties if not existent):

     - `ct page-prop-set -p -l "label1" -l "label2" "SpaceName:Parent Title"
       "SpaceName:Some Title" PageProp:='Content'`

       Additionally adds label "label1" and "label2"

     ct page-prop-set -p "SpaceName:Parent Title"
    """

    confluence = config.getConfluenceAPI()
    first = True

    PROPSET = re.compile(r'^(.*?)([:+-])=(.*)$')
    opmap = {
        ':': 'replace',
        '+': 'add',
        '-': 'remove',
    }

    files = []

    # handle propset
    propset = {}
    for prop in config['propset']:
        if prop.endswith('--'):
            propset[prop[:-2]] = 'delete'
        else:
            m = PROPSET.match(prop)
            if m:
                (name, op, value) = m.groups()
                if name not in propset:
                    propset[name] = {}

                _prop = propset[name]
                op = opmap[op]
                if op not in _prop:
                    _prop[op] = value
                else:
                    if not isinstance(_prop[op], list):
                        _prop[op] = [ _prop[op] ]
                    _prop[op].append(vlaue)

            else:
                files.append(prop)

    if not config.get('cql') or config.get('cql') == '-':
        files = ['-']
        config['cql'] = None

    # handle filenames
    input = ''
    for filename in files:
        if input != '':
            input += "---\n"
        if filename == '-':
            input += sys.stdin.read()
        else:
            with open(filename, 'r') as f:
                input += f.read()

    documents = []
    # parse yaml multidocument
    if input:
        for d in yaml.safe_load_all(input):
            if isinstance(d, list):
                documents += d
            else:
                documents.append(d)

    if len(propset):
        document = {
            'page': config['cql'],
            'pagePropertiesEditor': propset,
        }
        if config.get('parent', None):
            document['parent'] = config['parent']

        documents = [document] + documents
    else:
        if config['cql']:
            if len(documents) == 1:
                documents = [ {
                    'page': config['cql'],
                    'pagePropertiesEditor': documents[0]
                } ]

            if config.get('parent', None):
                documents[0]['parent'] = config['parent']

    labels = config.get('label', [])
    if labels is None:
        labels = []

    for doc in documents:
        if config.get('parent'):
            if 'parent' not in doc:
                doc['parent'] = config.get('parent')

        if 'pagePropertiesEditor' not in doc:
            if 'pageProperties' in doc:
                doc['pagePropertiesEditor'] = {}

                if isinstance(doc['pageProperties'], list):
                    _order = []
                    for e in doc['pageProperties']:
                        for k,v in e.items():
                            _order.append(k)
                            doc['pagePropertiesEditor'][k] = {'replace': v}
                    doc['pagePropertiesOrder'] = _order
                else:
                    for k,v in doc['pageProperties']:
                        doc['pagePropertiesEditor'][k] = {'replace': v}

                del doc['pageProperties']

        if 'page' not in doc and 'pages' not in doc:
            doc['page'] = "%s:%s" % (doc['spacekey'], doc['title'])

        doc_labels = doc.get('labels', doc.get('label', []))
        if not isinstance(doc_labels, list):
            doc_labels = [ doc_labels ]

        for result in confluence.setPageProperties(doc):
            if isinstance(result['page'], dict):
                print("created {spacekey}:{title} ({id})".format(id=result['result']['id'], **result['page'])
            else:
                print("updated {spacekey}:{title} ({id})".format(**(result['page'].dict('spacekey', 'title', 'id'))))

            _labels = labels+doc_labels
            if len(_labels):
                confluence.addLabels(result['result']['id'], _labels)



import contextlib, re
from pystache import Renderer
from os.path import dirname
from .myquery import MyQuery
from .util import get_list_data
from .page import Page
from lxml import etree

from lxml.etree import XMLSyntaxError

import logging
log = logging.getLogger('confluence-tool.storage_editor')

class StorageEditor:

    def __init__(self, confluence=None, templates=None, partials=None, actions=None):
        self.templates = templates
        self.partials = partials

        self.actions = get_list_data(actions)
        self.confluence = confluence
        self.renderer = Renderer(
            search_dirs = "{}/templates".format(dirname(__file__)),
            file_extension = "mustache",
            )


    def edit(self, content):
        if isinstance(content, Page):
            page = content
        else:
            page = None

        if isinstance(content, (dict, Page)):
            if content.get('body'):
                content = content['body']['storage']['value']

        try:
            Q = self.begin_edit(content)

        except XMLSyntaxError as e:
            if page is None: raise

            log.debug("error: %s", e)

            if re.match(r"Entity '.*?' not defined", str(e)):
                log.debug('matched')
                raise StandardError("There are double-escaped HTML entities in page %(spacekey)s:%(title)s" % page)
            else:
                raise

        for action in self.actions:
            if 'data' in action:
                if 'template' in action:
                    content = self.renderer.render_name(action['template'], action['data'])
                else:
                    content = self.renderer.render(action['content'], action['data'])
            else:
                content = action.get('content')

            if content is not None and 'type' in content:
                if content['type'] == 'wiki':
                    content = self.confluence.convertWikiToStorage(content)

            method = action.get('action', 'html')

            log.debug("content for %s: %s", method, content)

            selection = Q(action['select'])
            log.debug("selection: %s", selection)

            #import rpdb2 ; rpdb2.start_embedded_debugger('foo')

            if content is None:
                getattr(selection, method)()
            else:
                getattr(selection, method)(content)

            log.debug("edited: %s", str(Q))

        return self.end_edit()


    def begin_edit(self, content=None):
        if content is None:
            content = self.content

        self.pyquery = storage_query(content)

        return self.pyquery


    FIRST_OPENING_TAG = re.compile(r'^<root[^>]*>')
    LAST_CLOSING_TAG  = re.compile(r'</root>$')

    def end_edit(self, pyquery=None):
        if pyquery is None:
            pyquery = self.pyquery

        #import rpdb2 ; rpdb2.start_embedded_debugger('foo')

        root = pyquery.root.getroot()

        data = str(pyquery.__class__([x for x in root]))
        return data


def edit(content):
    storage_editor = StorageEditor()
    pyquery = storage_editor.begin_edit(content)

    def end_edit(*args):
        return storage_editor.end_edit()

    pyquery.end_edit = end_edit
    return pyquery


def storage_query(content):
    namespaces = {
        'ac': 'http://www.atlassian.com/schema/confluence/4/ac',
        'ri': 'http://www.atlassian.com/schema/confluence/4/ri',
    }

    return MyQuery(content, parser='xml', namespaces=namespaces)



"""
# Confluence Tool

Confluence Tool is for doing batch operations in confluence.

## Getting Started

For configuration, please run "confluence-tool -b BASE_URL config"

"""
from argdeco import CommandDecorator, arg, mutually_exclusive, group

command = CommandDecorator(
    arg('-c', '--config',      help="Configuration name", default='default'),
    arg('-C', '--config-file', help="Configuratoin file (default ~/.confluence-tool.yaml)"),
    arg('-b', '--baseurl',     help="Confluence Base URL, e.g. http://example.com/confluence"),
    arg('-u', '--username',    help="username for logging in (if not present, tried to read from netrc)"),
    arg('-p', '--password',    help="password for logging in (if not present, tried to read from netrc)"),
    arg('-d', '--debug',       action="store_true", help="get more information on exceptions"),
)

arg_cql = positional_arg_cql = arg('cql', help="SPACE:title, pageID or CQL, run '%(prog)s cql -h' for more help")
optarg_cql = positional_optarg_cql = arg('cql', nargs="?", help="SPACE:title, pageID or CQL, '%(prog)s cql -h' for more help")

arg_expand  = arg('-e', '--expand', help="values to expand")
arg_filter  = arg('-f', '--filter', help="page property filter run '%(prog)s page-prop-filtering -h' for more help")
arg_state   = arg('-s', '--state', help="get all pages for corresponding state '%(prog)s cw-states -h' for more help")
arg_status   = arg('-S', '--status', help="get all pages for corresponding status")
arg_write_format = arg('-w', '--write', help="format to write", choices=['format', 'yaml', 'json'], default="yaml")
arg_format  = arg('-F', '--format', help="format string for formatting the output.  May be either mustache or format string")
arg_parent  = arg('-p', '--parent', help="specify parent for a page, which might be created")
arg_message = arg('-m', '--message', help="add a note or message")
arg_add_label = arg('-l', '--label', action="append", help="add these labels to the page")
arg_label   = arg('-l', '--label', help="with label(s)")
arg_field   = arg('field', nargs="*", help='field to dump')

arg_parent = arg('-p', '--parent', help="specify parent for a page, which might be created")
#def arg_parent(parser, namespace, values, option_string=None):

@command('help-cql')
def cql_help(config):
    """How to query pages.

    You can pass [CQL] queries to many commands (indicated by parameter cql).
    In many queries you have to specify the ID, which originally needs an extra
    step to find it out.  So there is a convenience syntax for most common
    queries.

    [CQL]: https://developer.atlassian.com/confdev/confluence-server-rest-api/advanced-searching-using-cql

    Here you find a little translation of convenience syntax to CQL:

    - `<QUERY> ">>"` -> `ancestor = <ID of result of QUERY>`
    - `<QUERY> ">"`  -> `parent   = <ID of result of QUERY>`
    - `<SPACE> ":" <TITLE>` -> `space = <SPACE> and title = "<TITLE>"`
    - `":" <TITLE>`   -> `title = "<TITLE>"`
    - `<PAGE_ID>`  -> `ID = <PAGE_ID>`
    - `<PAGE_URI>` -> `ID = <page ID from PAGE_URI>`

    Examples:

     - `"IT:Some title>>"` to find all descendent pages of page with title
       "Some title" in space "IT"

     - `"IT:Some title"` to find page titled "Some title" in space "IT"

    """
    command['cql'].print_help()

@command('help-cw-states')
def comala_workflow_states(config):
    """How to filter page versions for specific state.

    If you want to get the Approved version of all pages, a simple
    `... and state = "Approved"` gives you only the pages currently in state
    Approved, but will not return the Approved version of pages, which are
    currently in a different state.

    The filter "--state <STATENAME>" will provide you the last version of the
    corresponding state.
    """
    command['cw-states'].print_help()

@command('help-page-prop-filtering')
def page_prop_filtering(config):
    """How to filter pages using page properties.

    Addionally to CQL queries, you can filter results (on client side) using
    page property filters.

    These filters are pretty simple, but helpful :)

    Page property filters are passes with `-F` command line option, and you can
    repeat this option to pass multiple filters, which all have to match against
    a page.
    l
    - `-F <NAME>==<VALUE>`  Value must be in property named `<NAME>`.  If
      property is a list, then this is true, if `<VALUE>` is one of the list
      values.  (works like labels in CQL).

    - `-F <NAME>!=<VALUE>`  Value must not be in property named `<NAME>`.

    - `-F '!<NAME>'`  Matches if property `<NAME>` is not present in page

    - `-F '<NAME>?'`  Matches if property `<NAME>` is present in page

    """

    command['page-prop-filtering'].print_help()

import pyaml, re, sys






@command('get',
    arg('url', help="url start with / or /rest/ will be prepended"),
    arg('--stream', help="get raw data", default=False, action="store_true"),
    arg('--progress', help="write out progress", default=False, action="store_true"),
    arg('--output-file', '-o', help="output file"),
    arg('params', nargs="*", help="parameters to pass to url"))
def get_method(config):
    """
    Get from a rest URL

    """

    PARAM = re.compile(r'(.*?)=(.*)')

    params = {}

    for item in config.get('params'):
        m = PARAM.search(item)
        if m:
            (name, value) = m.groups()
            params[name] = value

    url = config.get('url')

    output_file = config.get('output_file')
    if output_file and output_file != '-':
        outstream = open(output_file, 'wb')
    else:
        outstream = sys.stdout

    progress = config.get('progress')

    if 1:
        if not url.startswith('/'):
            url = '/rest/'+url

        confluence = config.getConfluenceAPI()

        result = confluence.get(url, stream=config.get('stream'), **params)
        if config.get('stream'):
            _len = 0
            for data in result.iter_content(chunk_size=1024*1024):
                outstream.write(data)

                _len += len(data)
                if progress:
                    sys.stderr.write("\r%s bytes" % _len)
        else:
            pyaml.p(result)



import logging, urllib
logger = logging.getLogger("confluence.page.props")

from os.path import dirname
#logger.setLevel(logging.DEBUG)

from .storage_editor import edit

from pyquery import PyQuery
from pystache import Renderer
from pprint import pprint

import re

DISPLAY_URL = re.compile(r'/display/([\w-]*)/(.*)')

def extract_data(elem, need_data=False):
    """Extracts data from confluence html page properties value <td>.

    Returns either a simple value in wiki represention or a complex data type
    with extraction of all containing data.
    """
    data = {}
    data['users'] = users = []
    data['refs']  = refs  = []
    data['links'] = links = []
    data['dates'] = dates = []
    data['mailAddresses'] = mailAddresses = []

    # if is list
    if elem.children().eq(0).is_('div.content-wrapper > ul, ul'):
        value = [ extract_data(PyQuery(e)) for e in elem.find('li') ]
        logger.debug("list value: %s", value)

    # if is dictionary-like table
    elif elem.children().eq(0).is_('div.table-wrapper, table'):
        value = {}
        for e in elem.find('table th'):
            if e.next_all().not_('td'):
                continue

            value[e.text().strip()] = extract_data(e.next_all())

    else:
        # transform usernames, refs and external links to wiki code
        d = PyQuery("<div>"+elem.html()+"</div>")

        for a in d('a'):
            _a = d(a)
            if _a.hasClass('confluence-userlink'):
                username = _a.attr('data-username')
                users.append(dict(username=username))
                _a.html("[~%s]" % username)

            elif _a.attr('href').startswith('mailto:'):
                address = _a.attr('href')[7:]
                mailAddresses.append(address)
                _a.html(address)

            elif _a.hasClass('external-link'):
                href = _a.attr('href')
                caption = _a.text()
                if caption != href:
                    _a.html("[%s|%s]" % (caption, href))
                    links.append(dict(caption=caption, href=href))
                else:
                    _a.html("[%s]" % href)
                    links.append(dict(href=href))
            else:
                #print _a.attr('href')
                m = DISPLAY_URL.search(_a.attr('href'))
                if m:
                    (space, title) = m.groups()
                    title = urllib.unquote(title).decode('utf8').replace(u"+", u" ")
                    _a.html("[{}:{}]".format(space, title))
                    refs.append(dict(space=space, title=title))

        for t in d('time'):
            _t = d(t)
            _time = _t.attr('datetime')
            dates.append(_time)
            _t.html(_time)

        value = d(':root').text().strip()

        logger.info("value: %s", value)
    if need_data:
        return dict(
            value=value, html=d.show(), users=users,
            refs=refs, links=links, mailAddresses=mailAddresses,
        )
    else:
        return value


def get_page_properties(html, need_html=False, need_data=False, properties=None, **kwargs):
    d = PyQuery(html)
    d('script').remove()
    th = d("div[data-macro-name=details] > div.table-wrap > table > tbody > tr > th")
    for elem in th:
        _th = PyQuery(elem)
        key = _th.text().strip()

        if properties is not None:
            if key not in properties:
                continue

        if not need_data and need_html:
            yield (key, _th.next().html())
        else:
            yield (key, extract_data(_th.next(), need_data=False))


class PagePropertiesEditor:

    def __init__(self, pagePropertiesEditor, templates={}, confluence=None, pagePropertiesOrder=None, **kwargs):
        self.editor = pagePropertiesEditor
        self.templates = templates
        self.confluence = confluence

        self.order = None
        if 'order' in kwargs:
            self.order = kwargs['order']
        if pagePropertiesOrder is not None:
            self.order = pagePropertiesOrder

        self.renderer = Renderer(
            search_dirs = "{}/templates".format(dirname(__file__)),
            file_extension = "mustache",
            )

        for t in ['user', 'page', 'link', 'list', 'value']:
            self.renderer.load_template(t)

        self.userkeys = {}


    def userkey(self, name):
        "resolve username to userkey"

        if name not in self.userkeys:
            self.userkeys[name] = self.confluence.getUser(name)['userKey']
        return self.userkeys[name]

    ELEM = re.compile(r'''(?: \[ (
         ~(?P<user>[^\]]*?)
        | (?P<spacekey>\w+):(?P<title>[^\]]*?)
        | (?P<link>[^\]]*?)
        ) \]
        | (?P<datetime>\d\d\d\d-\d\d-\d\d(?:\s\d\d:\d\d(?::\d\d)?)?)
        )
        ''', re.VERBOSE)


    def get_storage(self, key, value, templates=None):
        "get storage representation for value using templates"

        def render(name, **context):
            # try first, if there is a name-specific template
            for _name in ["{}-{}".format(key, name), name]:
                # first try local templates
                if _name in templates:
                    return self.renderer.render(templates[_name], context)
                # try editor templates
                elif _name in self.templates:
                    return self.renderer.render(self.templates[_name], context)

            # finally return the default template
            return self.renderer.render_name(name, context)

        if isinstance(value, list):
            values = [ {'value': self.get_storage(key, v, templates).strip() } for v in value ]
            return render('list', list=values)


        def replacer(m):
            d = m.groupdict()
            if d['user']:
                return render('user', userkey=self.userkey(d['user']))
            if d['spacekey']:
                return render('page', **d)
            if d['link']:
                if '|' in d['link']:
                    caption, href = d['link'].split('|')
                else:
                    caption, href = (d['link'],)*2
                return render('link', caption=caption, href=href)
            if d['datetime']:
                return render('datetime', datetime=d['datetime'])

        value = self.ELEM.sub(replacer, value)

        return render('value', value=value.strip())


    def edit_prop(self, key, data, action):

        if not isinstance(action, dict):
            return self.get_storage(key, action, {})

        if 'replace' in action:
            data = action['replace']

        if 'remove' in action:
            values = action['remove']
            if not isinstance(values, list):
                values = [ values ]

            if not isinstance(data, list):
                if values[0] == data:
                    data = []
                else:
                    logger.debug("replace(%s): value %s not present", key, values[0])
            else:
                for val in values:
                    if val in data:
                        data.remove(val)

        if 'add' in action:
            values = action['add']
            if not isinstance(values, list):
                values = [ values ]
            if not isinstance(data, list):
                if not data:
                    data = []
                else:
                    data = [ data ]
            data += values

        return self.get_storage(key, data, action.get('templates', {}))


    def edit(self, page=None):
        updated_keys = []
        if page is not None:
            content = page['body']['storage']['value']
            pageProperties = page['pageProperties']
        else:
            with open(dirname(__file__)+'/templates/page-props.html', 'r') as f:
                content = f.read()
            pageProperties = {}

        editor = edit(content)
        s = "ac|structured-macro[ac|name=details] > ac|rich-text-body > table > tbody > tr"

        for row in editor(s):
            key = editor(row).find('th').text().strip()

            if key not in self.editor: continue

            updated_keys.append(key)

            action = self.editor[key]
            if action == 'delete':
                editor(row).remove()
                continue

            data = pageProperties.get(key, '')
            html_data = self.edit_prop(key, data, action)
            x = editor(row).find('td')
            #import rpdb2 ; rpdb2.start_embedded_debugger('foo')

            x.html(html_data)

        selector = "ac|structured-macro[ac|name=details] table tbody"
        order = self.order
        if order is None:
            order = self.editor.keys()

        for key in order:
            action = self.editor[key]

            if action == 'delete': continue
            if key not in updated_keys:
                html_data = self.edit_prop(key, '', action)
                tr = "<tr><th>{}</th><td>{}</td></tr>".format(key, html_data)
                editor(selector).append(tr)

        return editor.end_edit()



"""
This module provides the main API of Confluence.  You can either import confluence_cli for having a confluence CLI or you import the API.
"""

from .confluence_api import ConfluenceError, ConfluenceAPI
from .cli import command, arg, main



from .cli import command, arg
import pyaml

@command('config',
    arg('-u', '--update', action="store_true", help="show data unless this specified"),
    arg('--update-password', action="store_true", help="unless specified, password will not be updated if not needed"),
    arg('--show-password', action="store_true", help="unless specified, password will not be shown"),
    )
def config(config):
    """Get or set configuration.

    For configuring the default base

    """

    if config.get('update') or config.get('update_password'):
        config.setConfig(config.get('update_password'))
        return

    if config.get('baseurl'):
        config.setConfig()
    else:
        cfg = config.getConfig()

        if cfg.get('show_password') and not config.get('show_password'):
            cfg['password'] = '******'

        pyaml.p(cfg)



#!/usr/bin/env python

from distutils.core import setup

setup(
    name='confluence-tool',
    version='0.0.18',
    description='Confluence API and CLI',
    author='Kay-Uwe (Kiwi) Lorenz',
    author_email='kiwi@franka.dyndns.org',
    install_requires=[
            'requests',
            'keyring',
            'keyrings.alt',
            'html5print',
            'pyquery',
            'pyaml',
            'pystache',
            'argdeco'
        ],
    url='https://github.com/klorenz/python-confluence-tool',
    packages=['confluence_tool', 'confluence_tool.cli', ],
    package_data = {
        'confluence_tool': ['templates/*.mustache', 'templates/*.html']
    },
    include_package_data = True,

    entry_points = dict(
        console_scripts = [
            'ct = confluence_tool:main',
            'confluence-tool = confluence_tool:main',
        ]
    ),
    license="MIT",
    )
